# moenlayout.tcl
#
#	Implements the Moen layout algorithm for graphical display of trees.
#
# Copyright (c) 1998 Clif Flynt 
#
# This program is free for use in non-commercial applications, provided 
# that the copyright information is retained.  It may be licensed from
# Flynt Consulting Services for use in commercial applications.    
#
#

#
# Placement algorithm
# 1) Go through tree: if no height/width component in a node, 
#    build a scratch canvas, render the node, generate a bbox, and calc ht/wd
#
# 2) Run Moen algorithm 
#    The Moen algorithm calculates a set of bounding polylines, and
#    merges these to create a mostly aesthetic view of the tree.
#
# 3) Draw Nodes


package provide moenlayout 1.0

namespace eval moenlayout {
namespace export moenLayout 
variable namespaceparent [namespace parent]

################################################################
# proc moenLayout {t} --
#   The top entry point for the Moen Node placement algorithm
# Arguments
#   The top node in the tree to position
# Results
#   Sets Xoffset and Yoffset for all nodes below t
# 
proc moenLayout {t} {
    variable namespaceparent 
    #
    # set node size, if not set

    if {[$t get height] == 0} {
	eval ${namespaceparent}::CanvasTreeSetSize $t
    }

    # Get childlist, process kids first

    set clst [$t childList]

    foreach c $clst {
	moenLayout $c
    }

    # If this node had kids, attach them, else layout as a leaf node.

    if {[llength $clst] > 0} {
	attach_parent $t [join $t]
    } else {
	layout_leaf $t
    }
}
################################################################
# proc layout_leaf {t} --
# Straight translation & steal from Moen
#  Sets the poly-lines to define a leaf
# 
# Arguments
#     t:	A leaf node
# Results
#     Sets the polylines for top&sides
# 
proc layout_leaf {t} {

    set width [$t get width]
    set height [$t get height]
    set xpad [$t get xpad]
    set ypad [$t get ypad]

puts "$t: xpad : $xpad ypad: $ypad"
    if {[string match [string tolower [$t get layout]] "horizontal"]} {
	$t set -node contour.upper.tail [list [expr $width + (2* $xpad)] 0]
	$t set -node contour.upper.head [$t get contour.upper.tail]
	$t set -node contour.lower.tail [list 0 [expr -$height - (2*$ypad)]]
	$t set -node contour.lower.head [concat [expr $width + (2*$xpad)] 0\
	  [$t get contour.lower.tail]]
    } else {
	$t set -node contour.upper.tail [list 0 [expr $height + (2* $ypad)]]
	$t set -node contour.upper.head [$t get contour.upper.tail]
	$t set -node contour.lower.tail [list [expr $width + (2*$xpad)] 0]
	$t set -node contour.lower.head [concat 0 [expr $height + (2*$ypad)]\
	  [$t get contour.lower.tail]]
    }
}

################################################################
# proc attach_parent {t h}
#     Attaches a parent to a set of child nodes.
#
# Arguments
#   t:	The parent node
#   h:  the aggregate size of the children nodes, 
#       height for horizontal layouts
#       width for vertical layouts
#
# Results
#   Resets contours and setx Xoffset and Yoffset fo primary child 
# relative to parent

proc attach_parent {t h} {

    set height [$t get height]
    set width [$t get width]
    set xpad [$t get xpad]
    set ypad [$t get ypad]

    set child [lindex [$t childList] 0]

    if {[string match [string tolower [$t get layout]] "horizontal"]} {
    set x [expr $xpad + [$t get parent_distance]]
    set y2 [expr ($h - $height) / 2- $ypad]
    set y1 [expr $y2 + $height + (2 * $ypad) - $h]
    $child set -node Xoffset [expr $x + $width]
    $child set -node Yoffset $y1
    $t set -node contour.upper.head [concat [$t get width] 0 $x $y1 [$t get\
      contour.upper.head]]

    $t set -node contour.lower.head [concat [$t get width] 0 $x $y2 [$t get\
      contour.lower.head]]

    } else {
    set y [expr $ypad + [$t get parent_distance]]

    set x1 [expr  ($width - $h)/2  +  $xpad]  ;#  Works 1 - LOOKS GOOD -
    set x2 [ expr $x1 - $width + $h - (2* $xpad)]

    $child set -node Xoffset $x1
    $child set -node Yoffset [expr $y + $height]
    $t set -node contour.upper.head [concat 0 [$t get height] [expr $x2] $y [$t get\
      contour.upper.head]]

    $t set -node contour.lower.head [concat 0 [$t get height ] [expr +$x1] $y [$t get\
      contour.lower.head]]

    }

}

################################################################
# proc join {t} --
# Joins a set of polylines - Parent & N children
# 
# Arguments
#   t:	The parent node.
# Results
# 
proc join {t} {
    set clst [$t childList]

    set ch0 [lindex $clst 0]
    set clst [lreplace $clst 0 0]

    # copy contours from first child to parent 

    foreach side [list upper lower] {
	foreach end [list head tail] {
	    $t set -node contour.$side.$end [$ch0 get contour.$side.$end]
	}
    }

    if {[string match [string tolower [$t get layout]] "horizontal"]} {
    set h [set sum [expr [$ch0 get height] + (2 * [$ch0 get ypad])]]
    } else {
    set w [set sum [expr [$ch0 get width] + (2 * [$ch0 get xpad])]]
    }

    set yoffset [$ch0 get ypad]
    set x [expr ([$t get xpad]) + [$t get parent_distance] + [$t get width]]

    foreach c $clst {
	set d [mergeContour $t $c]
        if {[string match [string tolower [$t get layout]] "horizontal"]} {
	$c set Yoffset [expr $d +$h]
	$c set Xoffset 0
	set h [expr [$c get height] + (2 * [$c get ypad])]
	incr sum [expr $d+$h]
	} else {
	$c set Xoffset [expr $d +$w]
	$c set Yoffset 0
	set w [expr [$c get width] + (2 * [$c get xpad])]
	incr sum [expr $d+$w]
	}

    }
    return $sum
}


################################################################
# proc mergeContour {c1 c2} --
#  The translation of merge()
#  Merges two polyline sets
#
# Arguments
#   c1:	node for upper set of polylines
#   c2:	node for lower set of polylines
# Results
#   Splits off to Vertical or Horizontal versions of code.
#   For trees laid out in vertical or horizontal fashion.

proc mergeContour {c1 c2} {
  if {[string match [string tolower [$c1 get layout]] "horizontal"]} {
      mergeContourHorizontal $c1 $c2
  } else {
      mergeContourVertical $c1 $c2
  }
}

################################################################
# proc mergeContourHorizontal {c1 c2} --
#   Horizontal layout merge function
#  see mergeContour
# 
# 
proc mergeContourHorizontal {c1 c2} {

    set upper [$c1 get contour.lower.head]
    set lower [$c2 get contour.upper.head]

    set x [set y [set total 0]]

    set ldy [set udy 1]
    set ldx [set udx 0]

    set ulen [llength $upper]
    set llen [llength $lower]

    while {($udy <= $ulen) &&($ldy <= $llen)} {

	set d [offset $x $y [lindex $lower $ldx] [lindex $lower $ldy]\
	  [lindex $upper $udx] [lindex $upper $udy]]
	incr y $d
	incr total $d

	if {($x + [lindex $lower $ldx]) < [lindex $upper $udx]} {
	    incr y [lindex $lower $ldy]
	    incr x [lindex $lower $ldx]
	    incr ldx 2

	    incr ldy 2

	} else {
	    incr y -[lindex $upper $udy]
	    incr x -[lindex $upper $udx]
	    incr udx 2

	    incr udy 2

	}
    }
    incr ldy -1

    incr udy -1

    if {$llen > $ldy} {
	set b [bridge $c1 contour.upper.tail 0 0 [lrange $lower $ldy end] $x $y]
	if {[llength $b] > 2} {
	    $c1 set -node contour.upper.tail [$c2 get contour.upper.tail]
	} else {
	    $c1 set -node contour.upper.tail $b
	}
	$c1 set -node contour.lower.tail [$c2 get contour.lower.tail]
    } else {
	set b [bridge $c2 contour.lower.tail $x $y [lrange $upper $udy end] 0 0]
	if {[llength $b] <= 2} {
	    $c1 set -node contour.lower.tail $b
	}
    }
    $c1 set contour.lower.head [$c2 get contour.lower.head]
    return $total
}

################################################################
# proc mergeContourVertical {c1 c2} --
#   Vertical layout merge function
#  see mergeContour
# 
# 
proc mergeContourVertical {c1 c2} {

    set upper [$c1 get contour.lower.head]
    set lower [$c2 get contour.upper.head]

    set x [set y [set total 0]]

    set ldy [set udy 1]
    set ldx [set udx 0]

    set ulen [llength $upper]
    set llen [llength $lower]

    while {($udy <= $ulen) &&($ldy <= $llen)} {

	set d [offset $y $x [lindex $upper $udy] [lindex $upper $udx] \
	  [lindex $lower $ldy] [lindex $lower $ldx] ]

	incr x $d
	incr total $d

	if {($y + [lindex $lower $ldy]) > [lindex $upper $udy]} {
	    incr y -[lindex $lower $ldy]
	    incr x -[lindex $lower $ldx]
	    incr ldx 2
	    incr ldy 2

	} else {
	    incr y [lindex $upper $udy]
	    incr x [lindex $upper $udx]
	    incr udx 2

	    incr udy 2

	}
    }
    incr ldy -1

    incr udy -1

    if {$llen > $ldy} {
	set b [bridge $c1 contour.upper.tail 0 0 [lrange $lower $ldy end] $x $y]
	if {[llength $b] > 2} {
	    $c1 set -node contour.upper.tail [$c2 get contour.upper.tail]
	} else {
	    $c1 set -node contour.upper.tail $b
	}
	$c1 set -node contour.lower.tail [$c2 get contour.lower.tail]
    } else {
	set b [bridge $c2 contour.lower.tail $x $y [lrange $upper $udy end] 0 0]
	if {[llength $b] <= 2} {
	    $c1 set -node contour.lower.tail $b
	}
    }
    $c1 set contour.lower.head [$c2 get contour.lower.head]
    return $total
}


################################################################
# proc bridge {node lineID x1 y1 line2 x2 y2} --
#     Bridges two sets of polylines
# Arguments
#     node:	The node which will have a contour modified.
#     lineID:	The name of the contour to change
#     x1:	A merge point
#     y1:	A merge point
#     line2	The second set of polylines
#     x2:	A merge point
#     y2:	A merge point
# Results
#     Modifies the $lineID controur of $node.
#     Creates a bridge between two sets of polylines.
# 

proc bridge {node lineID x1 y1 line2 x2 y2} {
  if {[string match [string tolower [$node get layout]] "horizontal"]} {
      bridgeHorizontal $node $lineID $x1 $y1 $line2 $x2 $y2
  } else {
      bridgeVertical $node $lineID $x1 $y1 $line2 $x2 $y2
  }
}

################################################################
# proc bridgeHorizontal {node lineID x1 y1 line2 x2 y2} --
#   see bridge
# 
proc bridgeHorizontal {node lineID x1 y1 line2 x2 y2} {

    set line1 [$node get $lineID]

    set dx [expr $x2 + [lindex $line2 0] - $x1]

    if {[lindex $line2 0] == 0} {
	set dy [lindex $line2 1]
    } else {
	set s [expr $dx * [lindex $line2 1]]
	set dy [expr $s / [lindex $line2 0]]
    }
    set r [concat $dx $dy [lrange $line2 2 end]]
    $node set lineID [concat [lindex $line1 0] [lindex $line1 1] 0 [expr $y2+\
      [lindex $line2 1] - $dy -$y1] $r]
    return $r
}

################################################################
# proc bridgeVertical {node lineID x1 y1 line2 x2 y2} --
#   see bridge
# 
proc bridgeVertical {node lineID x1 y1 line2 x2 y2} {

    set line1 [$node get $lineID]

    set dy [expr $y2 + [lindex $line2 1] - $y1]

    if {[lindex $line2 1] == 0} {
	set dx [lindex $line2 0]
    } else {
	set s [expr $dy * [lindex $line2 0]]
	set dx [expr $s / [lindex $line2 1]]
    }
    set r [concat $dx $dy [lrange $line2 2 end]]
    $node set lineID [concat [lindex $line1 0] [lindex $line1 1] [expr $x2+\
      [lindex $line2 0] - $dx -$x1] 0 $r]
    return $r
}

################################################################
# proc offset {p1 p2 a1 a2 b1 b2} --
#     Check the offsets for a set of points in polylines
# Arguments
#  p1, p2  	Test coords
#  a1, a2	Coordinates on line 1
#  b1, b2	Coordinates on line 2
# Results
#  Returns an offset if the lines need to be separated.
# 
proc offset {p1 p2 a1 a2 b1 b2} {
    if {($b1 <= $p1) || (($p1+$a1) <= 0)} {
	return 0
    }

    set t [expr $b1 * $a2 - $a1 * $b2]
    if {$t > 0} {
	if {$p1 < 0} {
	    set s [expr $p1 * $a2]
	    set d [expr ($s/$a1) - $p2]
	} else {
	    if {$p1 > 0} {
		set s [expr $p1 * $b2]
		set d [expr ($s/$b1) - $p2]
	    } else {
		set d [expr -$p2]
	    }
	}
    }\
    elseif {$b1 < ($p1+$a1)} {
	set s [expr ($b1 - $p1) * $a2]
	set d [expr $b2 - ($p2 + ($s/$a1))]
    }\
    elseif {$b1 > ($p1+$a1)} {
	set s [expr ($a1 + $p1) * $b2]
	set d [expr ($s/$b1) - ($p2 +$a2)]
    } else {
	set d [expr $b2 - ($p2+$a2)]
    }
    if {$d > 0} {
	return $d
    } else {
	return 0
    }
}



}
