# tree.tcl --
#	This file implements a tree data structure for Tcl
#	programs.
#
# Copyright (c) 1998 Clif Flynt 
#
# This program is free for use in non-commercial applications, provided 
# that the copyright information is retained.  It may be licensed from
# Flynt Consulting Services for use in commercial applications.    
#
# If -test is only command line arg, then this is being run standalone
#    for a test.

set tree_version [lindex {$RCSfile: tree.tcl,v $ $Revision: 1.31 $} 4]

# This is the naming convention for the array in which the tree's are
#   implemented:
#
# The package includes one entry point that is used to define unique
#   names for each tree created.
#
# Tree_InternalArray(base) -  A value for creating unique tree names.
#
# The $Tree_InternalArray(base) value is appended to "tree" to form a
#   unique identifier for each tree.  This is refered to as the "primary"
#   in the rest of this discussion.
#
#  Each tree includes a private base value that is used to create unique
#    names for each node.  That index into the Tree_InternalArray is:
#
#  primary.base - contains next unique ID number for creating nodes.
#
# Nodes linkage indices are named as:
#
# Tree_InternalArray(primary.idNum.field)
#
#  treeID.nodeID.children  -
#      List of children for node nodeID. {treeID.nodeID treeID2.nodeID2...}
#  treeID.nodeID.parent  - Parent for node - treeID.nodeID
#
# Key/Value pairs are indices are named as:
#
#  treeID.nodeID.node.Key - An Key/Value pair for this node.
#  treeID.tree.Key - Key/Value pairs for entire treeID tree.
#

set Tree_InternalArray(base) 0


########################################################################
# proc TreeGetNameParts {name Base IdNum} --
#  Internal proc for splitting a name into parts
# Arguments:
#  name:	The name to split
#  Base:	The name of the tcl variable to receive the base name
#  IdNum:	The name of the tcl variable to receive the ID number
#
#  No valid Return, sets two variables in the calling proc.

proc TreeGetNameParts {name Base IdNum {validate 1}} {
    global Tree_InternalArray

    upvar $Base base
    upvar $IdNum idNum

    set namelst [split $name "."]
    set base [lindex $namelst 0]
    set idNum [lindex $namelst 1]

    if {$validate} {
        if {[array name Tree_InternalArray $base.base] == ""} {
  	    error "$base.base does not exist - did you treeCreate it?"
	}
    }
}

########################################################################
# proc treeCreate {} --
#    Create a root node for a tree
# Arguments:
#
# Results:
#	Creates a new root node in Tree_InternalArray
#	Initializes a unique counter for this tree.
#       Returns the name of the root node for use as a handle
# 
proc treeCreate {} {
    global Tree_InternalArray


    #
    # Create the unique name for this tree.

    set primaryName "tree$Tree_InternalArray(base)"
    incr Tree_InternalArray(base)

    #
    # Can't have two trees with the same name.
    # This should never happen, but just in case...

    if {[array name Tree_InternalArray $primaryName.base] != ""} {
	error "$primaryName already exists"
    }

    #
    # Set the initial unique ID for nodes in this tree.
    #
    set Tree_InternalArray($primaryName.base) 0


    #
    # Create the root node
    #
    set rootNode [TreeCreateNode $primaryName.base]

    #
    # Looks good, ship it!
    #
    return $rootNode
}

########################################################################
# proc treeCreateChild {parent}  --
#   Creates a child of a given parent
#
# Arguments:
#	parent:	The parent node to contain this new node
# Results:
#	Creates a new node.
#	Adds the child to the parent node.
#	Returns the handle for the new node
#
proc treeCreateChild {parent} {
    global Tree_InternalArray


    TreeGetNameParts $parent primary idNum

    set child [TreeCreateNode $primary.base]


    TreeAddChild $parent $child
    return $child
}

########################################################################
# proc TreeCreateNode {rootname} --
#  Create a new node (unattached) in this tree 
# Arguments: 
#    basename:  The name of a node in the tree
#
# Results:
#    Creates a new node.
#    Returns a handle to use to identify this node.
#
proc TreeCreateNode {rootname} {
    global Tree_InternalArray

    # The existence of the tree is checked in TreeGetNameParts
    
    TreeGetNameParts $rootname primary idNum

    set val [incr Tree_InternalArray($primary.base)]
    set childname "$primary.$val"
    set Tree_InternalArray($childname) ""
    return $childname
}

########################################################################
#  proc TreeAddChild {parent child} --
#    Adds a node to the child list of the parent node, and sets child's
#    parent pointer to reflect the parent node.
#
# Arguments:
#	parent:	The parent node
#	child:	New child node being added to the parent.
#
# Results:
#	Adds the child and updates lists & pointers
#	No valid return.
#
proc TreeAddChild {parent child} {
    global Tree_InternalArray


    if {[array name Tree_InternalArray $parent] == ""} {
	error "TreeAddChild: Parent $parent does not exist "
    }

    if {[array name Tree_InternalArray $child] == ""} {
	error "TreeAddChild: Child $child does not exist "
    }


    TreeGetNameParts $parent primary parentID

    TreeGetNameParts $child primary childID

    if {[info exists Tree_InternalArray($primary.$childID).parent]} {
	error "$child already has parent: \
	  $Tree_InternalArray($primary.$childID.parent)"
    }

    lappend Tree_InternalArray($primary.$parentID.children) $child

    set Tree_InternalArray($primary.$childID.parent) $parent
}

########################################################################
# proc treeDestroyTree {nodeName} --
#  Destroy a tree - including the root entry
# Arguments: 
#    nodeName:	The node to delete
#
# Results:
#    Deletes a node, and all associated Key/Value pairs.
#    Deletes all children if recurs is != 0
#
proc treeDestroyTree {nodeName} {
    global Tree_InternalArray


    TreeGetNameParts $nodeName primary idNum

    # delete all references to this array.

    set namelist [array name Tree_InternalArray $primary.*]
    foreach name $namelist {
	unset Tree_InternalArray($name)
    }

    return ""
}

########################################################################
# proc treeDeleteNode {nodeName {recurs 0}} --
#   Delete a node in a tree.  
#   If recurs is != 0, delete all children of this node.
# Arguments: 
#    nodeName:	The node to delete
#    ?recure?:	if set != 0, delete all this nodes children.
#
# Results:
#    Deletes a node, and all associated Key/Value pairs.
#    Deletes all children if recurs is != 0
#

proc treeDeleteNode {nodeName {recurs 0}} {
    global Tree_InternalArray


    if {[array name Tree_InternalArray $nodeName] == ""} {
	error "$nodeName does not exist - did you ever create it?"
    }
    TreeGetNameParts $nodeName primary idNum

    set Tree_InternalArray($nodeName) DEAD


    # delete the Key/value pairs for this node.

    set namelist [array name Tree_InternalArray $nodeName.node.*]
    foreach name $namelist {
	unset Tree_InternalArray($name)
    }

    #
    # Remove references to this node.
    #

    foreach node [array names Tree_InternalArray $primary.*.children] {
	if {[set pos [lsearch $Tree_InternalArray($node) $nodeName]] != -1} {
	    set Tree_InternalArray($node) [lreplace $Tree_InternalArray($node)\
	      $pos $pos]
	}
    }

    #
    # If recursive, delete this nodes children also.
    #

    if {$recurs} {
	if {[info exists Tree_InternalArray($primary.$idNum.children)]} {
	    foreach node $Tree_InternalArray($primary.$idNum.children) {
		treeDeleteNode $node $recurs
	    }
	}
    } else {
	# else, relink children to this node's parents
	if {[info exists Tree_InternalArray($primary.$idNum.children)]} {
	    foreach node $Tree_InternalArray($primary.$idNum.children) {
		TreeGetNameParts $node primary nodeId
		unset Tree_InternalArray($primary.$nodeId.parent)
		TreeAddChild $Tree_InternalArray($primary.$idNum.parent) $node
	    }
	}
    }

    #
    # If other deletes have been done, this may be a dead node,
    # with no parent.
    #
    if {[info exists Tree_InternalArray($primary.$idNum.parent)]} {
	unset Tree_InternalArray($primary.$idNum.parent)
    }

    unset Tree_InternalArray($nodeName)

    return ""
}

########################################################################
#  proc treeGetChildList {parent} --
#  Return the list of children a parent node points to.
#
# Arguments:
#	parent:	The parent node to get the childlist from.
#
# Results:
# 	Returns a list of child nodes.
#	Does not modify tree
#
proc treeGetChildList {parent} {
    global Tree_InternalArray


    TreeGetNameParts $parent primary parentID

    if {[array name Tree_InternalArray $parent] == ""} {
	error "TreeGetChildList: Parent $parent does not exist "
    }

    if {[array name Tree_InternalArray $primary.$parentID.children] == ""} {

	#        error "TreeGetChildList: Parent $parent has no children"
	return ""
    }

    return $Tree_InternalArray($primary.$parentID.children)
}

########################################################################
# proc treeGetSiblingList {node} --
# 	Return a list of siblings - the parents childList less self
#
# Arguments:
#	node	The node to return a list of sibs of.
#
# Results:
#	Returns a list of siblings
#	or empty string if no parent or no sibs.
#
proc treeGetSiblingList {node} {
    global Tree_InternalArray
    
    set fail [catch {treeGetParent $node} parent]
    if {$fail} {return ""}

    set childList [treeGetChildList $parent]
    set mypos [lsearch $childList $node ]
    set childList [lreplace $childList $mypos $mypos]
    return $childList
}

########################################################################
# proc treeGetParent {node} --
# 	Returns a parent of a node (if one exists)
#
# Arguments:
#	node	The child node to return the parent from
#
# Results:
#	Returns the parent node,
#	Or generates an error if node, or parent does not exist.
#
proc treeGetParent {node} {
    global Tree_InternalArray

    TreeGetNameParts $node primary parentID

    if {[array name Tree_InternalArray $node] == ""} {
	error "treeGetParent: Node $node does not exist "
    }

    if {![info exists Tree_InternalArray($primary.$parentID.parent)]} {
	error "treeGetParent: $node has no parent"
    }
    return $Tree_InternalArray($primary.$parentID.parent)
}

########################################################################
#  proc treeSetNodeValue {node key value} --
#	Sets the value of a key/value pair in a node.
#
# Arguments:
#	node:	The node to set the key/value pair in
#	key:	An identifier to set to a value
#	value:	The value for this identifier
#
# Results:
#	Sets an key/value pair in this node.
#	Returns the value
#
proc treeSetNodeValue {node key value} {
    global Tree_InternalArray


    TreeGetNameParts $node primary parentID

    if {[array name Tree_InternalArray $node] == ""} {
	error "treeSetNodeValue: Node $node does not exist "
    }

    set Tree_InternalArray($node.node.$key) $value
    return $value
}
########################################################################
#  proc treeGetNodeValue {node key} --
#	Returns a value from a key/value pair, if it has previously been set.
#
# Arguments:
#	node:	The node to get the key/value pair from
#	key:	The identifier to get a value for
#
# Results:
#	Returns the value from a previously set key/value pair
#
proc treeGetNodeValue {node key} {
    global Tree_InternalArray


    TreeGetNameParts $node primary idNum

    if {[array name Tree_InternalArray $node] == ""} {
	error "treeGetNodeValue: Node $node does not exist "
    }

    if {[array name Tree_InternalArray $node.node.$key] == ""} {
	error "treeGetNodeValue: Key $key does not exist "
    }

    return $Tree_InternalArray($node.node.$key)
}

########################################################################
#  proc treeGetNodeValueList {node} --
#	Returns a list of all key/value pairs.
#
# Arguments:
#	node:	The node to extract the list of key/value pairs from
#
# Results:
#	Returns a list of key/value pairs as {{key1} {value1} {key2} {value2}... }
proc treeGetNodeValueList {node} {
    global Tree_InternalArray


    TreeGetNameParts $node primary idNum

    if {[array name Tree_InternalArray $node] == ""} {
	error "treeGetNodeValueList: Node $node does not exist "
    }

    set lst ""

    if {[array name Tree_InternalArray $node.node.*] != ""} {
	set namelist [array name Tree_InternalArray $node.node.*]
	foreach name $namelist {
	    lappend lst [list $name $Tree_InternalArray($name)]
	}
    }

    return $lst
}

########################################################################
#  proc treeSetTreeValue {node key value} --
#	Sets the value of an key/value pair in a node.
#
# Arguments:
#	node:	A node in the Tree in which to set a key/value pair
#	key:	An identifier to set to a value
#	value:	The value for this identifier
#
# Results:
#	Sets an key/value pair in this node.
#	Returns the value
#
proc treeSetTreeValue {node key value} {
    global Tree_InternalArray


    TreeGetNameParts $node primary parentID

    set Tree_InternalArray($primary.tree.$key) $value
    return $value
}
########################################################################
#  proc treeGetTreeValue {node key} --
#	Returns a value from an key/value pair, if it has previously been set.
#
# Arguments:
#	node:	A node in the Tree from which to retrieve a key/value pair
#	key:	The identifier to get a value for
#
# Results:
#	Returns the value from a previously set key/value pair
#
proc treeGetTreeValue {node key} {
    global Tree_InternalArray


    TreeGetNameParts $node primary idNum

    if {[array name Tree_InternalArray $primary.tree.$key] == ""} {
	error "treeGetTreeValue: Key $key does not exist "
    }

    return $Tree_InternalArray($primary.tree.$key)
}

########################################################################
#  proc treeGetTreeValueList {node} --
#	Returns a list of all key/value pairs.
#
# Arguments:
#	node:	The node to extract the list of key/value pairs from
#
# Results:
#	Returns a list of key/value pairs as {{key1} {value1} {key2} {value2}... }
proc treeGetTreeValueList {node} {
    global Tree_InternalArray

    TreeGetNameParts $node primary idNum

    set lst ""

    if {[array name Tree_InternalArray $primary.tree.*] != ""} {
	set namelist [array name Tree_InternalArray $primary.tree.*]
	foreach name $namelist {
	    lappend lst [list $name $Tree_InternalArray($name)]
	}
    }

    return $lst
}

################################################################
# proc treeSearchKey {node key} --
#    Searches the appropriate tree for a node with key fields matching key
# Arguments
#   node:	A node in the tree to search
#   key:	The key to search for
# Results
#   Returns a list of node names.
# 
proc treeSearchKey {node key} {
    variable Tree_InternalArray

    TreeGetNameParts $node primary idNum

    set lst [array names Tree_InternalArray "$primary.*.node.$key"]
    
    set lst2 ""
    foreach n $lst {
        TreeGetNameParts $n tree id
	lappend lst2 "$tree.$id"
    }

    return $lst2
}

########################################################################
# proc treeGetRoot {node} --
#    Return the top parent of a node.
# Arguments: 
#    node:	The node to search from.
#
# Results:
#    Returns the base node that leads to a given node.  The ultimate parent.
#

proc treeGetRoot {node} {
    global Tree_InternalArray


    while {![catch {treeGetParent $node} n]} {
	set node "$n"
    }
    return $node

}


########################################################################
# proc treeDump {node ?recurs? ?spaces?} --
#    Dump the contents of a node.
# Arguments: 
#    node:	The node dump
#    recurs:	Will dump recursively if set to -r
#    spaces:	For internal use to set the indent level.
#
# Results:
#    returns a formatted list of array indices and values associated 
#    with this node, suitable for each list element to be output in a
#    semi-pretty format.
#

proc treeDump {node {recurs ""} {spaces ""}} {
    global Tree_InternalArray


    TreeGetNameParts $node primary idNum

    if {$spaces == ""} {
	set Tree_InternalArray(dump) ""
    }

    lappend Tree_InternalArray(dump) "$spaces NODE: $node - primary: $primary\
      - idNum: $idNum"

    if {[array name Tree_InternalArray $node] == ""} {
	error "treeGetNodeValueList: Node $node does not exist "
    }

    if {[info exists Tree_InternalArray($primary.$idNum.parent)]} {
	lappend Tree_InternalArray(dump) "$spaces  Parent:\
	  $Tree_InternalArray($primary.$idNum.parent)"
    } else {
	lappend Tree_InternalArray(dump) "$spaces  Parent: None"
    }

    if {[array name Tree_InternalArray $node.node.*] == ""} {
	lappend Tree_InternalArray(dump) "$spaces  treeGetNodeValueList: Node\
	  $node has no values"
    } else {
	set namelist [array name Tree_InternalArray $node.node.*]
	lappend Tree_InternalArray(dump) "$spaces  Values:"
	foreach name $namelist {
	    lappend Tree_InternalArray(dump) "$spaces    $name\
	      $Tree_InternalArray($name)"
	}
    }

    if {[info exists Tree_InternalArray($primary.$idNum.children)]} {
	lappend Tree_InternalArray(dump) "$spaces  Children:\
	  $Tree_InternalArray($primary.$idNum.children)"
	if {[string match $recurs "-r"]} {
	    set spaces "$spaces    "
	    foreach name $Tree_InternalArray($primary.$idNum.children) {
		treeDump $name -r $spaces
	    }
	}
    } else {
	lappend Tree_InternalArray(dump) "$spaces  Children: None"
    }
    return $Tree_InternalArray(dump)

}

if {([info exists argv]) && ([lindex $argv 0] == "-test" )} {
    proc TreeTest {} {

	# test treeCreate

	if {![string match [set parent1 [treeCreate]] "tree0.1"]} {
	    error "Failed to get expected return from treeCreate:  $parent1 !=\
	      tree0.1"
	}

	if {![string match [set parent2 [treeCreate]] "tree1.1"]} {
	    error "Failed to get expected return from treeCreate:  $parent2 !=\
	      tree1.1"
	}

	# test TreeCreateNode and AddChild

	if {![string match [set ch1 [TreeCreateNode $parent1]]\
	  "tree0.2"]} {
	    error "Failed to get expected return from treeCreateNode:  $ch1!=\
	      tree0.2"
	}

	if {![string match [set tmp [TreeAddChild $parent1 $ch1]]\
	  "tree0.1"]} {
	    error "Failed to get expected return from TreeAddChile:  $tmp !=\
	      tree0.1"
	}

	# test treeCreateChild, and add some more nodes for other testing.

	if {![string match [set ch2 [treeCreateChild $parent2]]\
	  "tree1.2"]} {
	    error "Failed to get expected return from treeChildCreate:  $ch2!=\
	      tree1.2"
	}

	if {![string match [set ch1_1 [treeCreateChild $ch1]] "tree0.3"]} {
	    error "Failed to get expected return from treeChildCreate: \
	      $ch1_1!= tree0.3"
	}

	set ch2_1 [treeCreateChild $ch2]

	set ch1_2 [treeCreateChild $ch1]
	set ch2_2 [treeCreateChild $ch2]

	# confirm that children know their parents.

	if {![string match [set tmp [treeGetParent $ch2]] "tree1.1"]} {
	    error "Failed to get expected return from treeChildCreate:  $tmp\
	      != tree.1"
	}

	# Confirm that childlists match expected lists.

	if {![string match "tree0.2" [treeGetChildList $parent1]]} {
	    error "Failed to match childlist ([treeGetChildList $parent1]) to\
	      tree0.2"
	}

	if {![string match "tree0.3 tree0.4"\
	  [treeGetChildList $ch1]]} {
	    error "Failed to match childlist ([treeGetChildList $ch1]) to\
	      \"tree0.3 tree0.4\""
	}

	if {![string match "tree1.2" [treeGetChildList $parent2]]} {
	    error "Failed to match childlist ([treeGetChildList $parent2]) to\
	      tree1.2"
	}

	# Check treeSetNodeValue - assign values to nodes, and compare

	foreach node [treeGetChildList $ch1] {
	    treeSetNodeValue $node $node.id $node.Value
	}

	foreach node [treeGetChildList $ch1] {
	    if {![string match [treeGetNodeValue $node $node.id]\
	      "$node.Value"]} {
		error "Set value to $node.Value and got back \
		  [treeGetNodeValue $node $node.id]"
	    }
	}

	set value [treeGetNodeValue $ch1_2 $ch1_2.id]
	set expectValue "tree0.4.Value"

	if {![string match $value $expectValue]} {
	    error "Failed to get a match in  treeGetNodeValue: $value !=\
	      $expectValue"
	}

	# Confirm treeGetNodeValueList works OK.

	set valueList [treeGetNodeValueList $ch1_2]
	set expectValueList {{tree0.4.node.tree0.4.id tree0.4.Value}}

	if {![string match $valueList $expectValueList]} {
	    error "Failed to get a match in  treeGetNodeValueList: $valueList\
	      != $expectValueList"
	}

	# Check treeSetTreeValue
	treeSetTreeValue $parent1 TreeKey1 TreeKeyVal1
	treeSetTreeValue $parent1 TreeKey2 TreeKeyVal2

	# Confirm that tree data is available to other nodes on tree.

	if {[treeGetTreeValue $ch1_2 TreeKey1] != "TreeKeyVal1"} {
	    error "Failed to get tree value for $ch1_2 : TreeKey1"
	}

	# Confirm that treeGetTreeValueList works as expected.

	set expectList {{tree0.tree.TreeKey1 TreeKeyVal1}\
	  {tree0.tree.TreeKey2 TreeKeyVal2}}
	set valueList [treeGetTreeValueList $ch1_1]
	if {![string match $valueList $expectList]} {
	    error "treeGetTreeValueList failed: $valueList != $expectList"
	}

	# Check treeDump for single node.

	set nodeDump [treeDump $ch1]
	set expectDump {{ NODE: tree0.2 - primary: tree0 - idNum: 2}\
	  {  Parent: tree0.1} {  treeGetNodeValueList: Node tree0.2\
	  has no values} {  Children: tree0.3 tree0.4}}

	if {![string match $nodeDump $expectDump]} {
	    puts "Expect: $expectDump"
	    puts "got:    $nodeDump"
	    error "Failed to get a match in  node dump of ch1."
	}

	# Check treeDump recursive option

	set parentDump [treeDump $parent1 -r]
	set expectDump {{ NODE: tree0.1 - primary: tree0 - idNum: 1}\
	  {  Parent: None} {  treeGetNodeValueList: Node tree0.1 has no\
	  values} {  Children: tree0.2} {     NODE: tree0.2\
	  - primary: tree0 - idNum: 2} {      Parent: tree0.1}\
	  {      treeGetNodeValueList: Node tree0.2 has no values}\
	  {      Children: tree0.3 tree0.4} {         NODE:\
	  tree0.3 - primary: tree0 - idNum: 3} {          Parent:\
	  tree0.2} {          Values:}\
	  {            tree0.3.node.tree0.3.id tree0.3.Value}\
	  {          Children: None} {         NODE: tree0.4 - primary:\
	  tree0 - idNum: 4} {          Parent: tree0.2}\
	  {          Values:} {            tree0.4.node.tree0.4.id\
	  tree0.4.Value} {          Children: None}}

	if {![string match $parentDump $expectDump]} {
	    puts "Expect: $expectDump"
	    puts "got:    $parentDump"
	    error "Failed to get a match in  recursive dump of parent."
	}

	# Check treeGetRoot

	if {![string match [set tmp [treeGetRoot $ch2_2]] "tree1.1"]} {
	    error "Failed to get expected return from treeGetRoot:  $tmp !=\
	      tree1.1"
	}

	# Set a random value, and create a new child node.
	# delete the node, 
	# confirm the node cannot be accessed for its value.
	# confirm that the node's child has its parent remapped.

	treeSetNodeValue $ch2_2 id2_2 value2_2

	if {![string match [set ch2_2_1 [treeCreateChild $ch2_2]]\
	  "tree1.5"]} {
	    error "Failed to get expected return from treeChildCreate: \
	      $ch2_2_1 != tree1.5"
	}

	set parent_2 [treeGetParent $ch2_2]

	treeDeleteNode $ch2_2

	if {1 != [catch {treeGetNodeValue $ch2_2 id2_2}]} {
	    error "treeGetNodeValue succeeded on a deleted node!"
	}

	# Confirm that parent is remapped for child of deleted node.

	set parent_D2 [treeGetParent $ch2_2_1]
	if {![string match $parent_D2 $parent_2]} {
	    error "Bad Parent Remapping: Deleted: $ch2_2, and child parent\
	      became: $parent_D2 instead of $parent_2"
	}

	# Delete the full tree.

	treeDestroyTree $parent1

	if {1 != [catch {treeGetTreeValueList $parent1}]} {
	    error "treeGetTreeValueList succeeded on a deleted tree!"
	}

	if {1 != [catch {treeGetTreeValueList $ch1}]} {
	    error "treeGetTreeValueList succeeded on node of a deleted tree!"
	}

	if {1 != [catch {treeGetTreeValueList $ch1_1}]} {
	    error "treeGetTreeValueList succeeded on node of a deleted tree!"
	}

	if {1 == [catch {treeGetTreeValueList $parent2}]} {
	    error "treeGetTreeValueList failed on an  undeleted tree!"
	}

	# confirm can still create trees.
	if {![string match [set parent1a [treeCreate]] "tree2.1"]} {
	    error "Failed to get expected return from treeCreate:  $parent1a\
	      != tree2.1"
	}
	puts "TreeTest OK -- No errors"
    }

    TreeTest
}
